#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HsPyLib-Clitt
   @package: clitt.core
      @file: preferences.py
   @created: Fri, 7 Jul 2023
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2023, HsPyLib team
"""
import os
from typing import Any

from hspylib.core.metaclass.singleton import AbstractSingleton
from hspylib.core.preconditions import check_argument, check_not_none
from hspylib.core.tools.dict_tools import get_or_default_by_key
from hspylib.core.tools.text_tools import ensure_startswith, environ_name


class Preferences(metaclass=AbstractSingleton):
    """General purpose bas class to provide preference management."""

    def __init__(self, prefix: str):
        self._prefix = prefix
        self._overrides = {}

    def __str__(self):
        # fmt: off
        return (
            f"Terminal UI Preferences{os.linesep}"
            f"{'-=' * 20}{os.linesep}"
            + os.linesep.join([f"|-{p[1:]:<16}: {getattr(self, p)}" for p in vars(self)])
            + f"{os.linesep}{'-=' * 20}{os.linesep}"
        )
        # fmt: on

    def __repr__(self):
        return str(self)

    def __getitem__(self, name: str):
        return self.get_preference(self._get_name(name))

    def __setitem__(self, name: str, value: Any):
        curr_val = getattr(self, name.replace(".", "_"))
        t1, t2 = type(curr_val), type(value)
        check_not_none(curr_val, f"Preference {name} does not exist")
        check_argument(t1 == t2, f"Preference {name} value must be of type {t1}, not {t2}")
        self._overrides[self._get_name(name)] = value

    def __iter__(self):
        return self._overrides.__iter__()

    def __len__(self):
        return len(self._overrides)

    def get_preference(self, name: str, default: Any = None) -> Any:
        """TODO
        :param name: TODO
        :param default: TODO
        """
        pref_name = self._get_name(name)
        value = get_or_default_by_key(self._overrides, pref_name, default)
        return os.environ.get(environ_name(pref_name), value)

    def _get_name(self, name: str) -> str:
        """TODO
        :param name: TODO
        """
        return ensure_startswith(name, self._prefix + '.')
